// DEVICE_IO Library

#include "propeller.h"

#define MASK(pin) (1 << (pin))
#define SHIFTR(val, cnt) ((val) >> (cnt))
#define SHIFTL(val, cnt) ((val) << (cnt))
#define BITON(val, pin) ((val) |= (MASK(pin)))
#define BITOFF(val, pin) ((val) &= ~(MASK(pin)))

#define _i2cACK     0
#define _i2cNAK     1

typedef enum {LSBFIRST, MSBFIRST} ShiftOutMode;
typedef enum {MSBPRE, MSBPOST, LSBPRE, LSBPOST} ShiftInMode;


const int PULSECNT = 1000;
unsigned long i2cDat, i2cClk, spiMOSI, spiMISO, spiCLK, spiSPEED;
unsigned char spiMODE;

static void             i2cStop ( void );
static void             i2cStart( void );
static unsigned char    i2cWrite( unsigned char i2cData );
static unsigned char    i2cRead ( unsigned char ackbit);

unsigned int BitSet(unsigned int value, unsigned char bit)
{
	BITON(value, bit);
	return value;
}


void BitClear(unsigned int value, unsigned char bit)
{
	BITOFF(value, bit);
}

void SetDIR(unsigned char pin, unsigned char state)
{
	state ? BITON(DIRA, pin) : BITOFF(DIRA, pin);
}


unsigned int BitRead(unsigned int value, unsigned char bit)
{
   return (value >> bit) & 1;
}

unsigned char BitWrite(unsigned int value, unsigned char bit, unsigned char state)
{
	state ? BITON(value, bit) : BITOFF(value, bit);
	return value;
}

void Out(unsigned char pin, unsigned char state)
{
   state &= 1;
	
	BITON(DIRA, pin);
   OUTA = BitWrite(OUTA, pin, state);
}

unsigned char In(unsigned char pin)
{
   BITOFF(DIRA, pin);
   return  (SHIFTR((INA & MASK(pin)), pin));
}


void Delay(unsigned int duration)
{
   unsigned int time;

   time = CNT;
   waitcnt(duration + time);
}

void PulseOut(unsigned char pin, unsigned char state, unsigned int duration)
{
   Out(pin, state); 
   Delay(duration);
   Out(pin, ~state);
}

void ShiftOut(unsigned char Dpin, unsigned char Cpin, unsigned int Value, ShiftOutMode Mode, unsigned char bits)
{
   unsigned char i;
   unsigned int val;
  	
   for(i = 0; i < bits; i++)
   {
      if (Mode == LSBFIRST)
      {
	      val = Value & 1;
         Value = Value >> 1;
      }
      else
      {
	      val = (Value >> (bits - 1)) & 1;
	      Value = Value << 1;
      }
     
      Out(Dpin, val);
      PulseOut(Cpin, 1, PULSECNT);
      Delay(PULSECNT);
   }
}

unsigned int ShiftIn(unsigned char Dpin, unsigned char Cpin, ShiftInMode Mode, unsigned char Bits)
{
   unsigned int Inbit;
   unsigned char i;
   unsigned int Value = 0;

	switch(Mode)
   {
	   case MSBPRE:
			for(i = 0; i < Bits; i++)
   		{
	 		 	Inbit = In(Dpin);
	  			Value = (Value << 1) | Inbit;
				PulseOut(Cpin, 1, PULSECNT);
            Delay(PULSECNT);
         }
	      break; 
      case MSBPOST:
	      for(i = 0; i < Bits; i++)
	      {
            PulseOut(Cpin, 1, PULSECNT);
	        	Inbit = In(Dpin);
	         Value = (Value << 1) | Inbit;
            Delay(PULSECNT);
	      }
	      break;
      case LSBPRE:
	      for(i = 0; i < Bits; i++)
	      {	 
	        	Inbit = In(Dpin);
	         Value = (Value >> 1) | (Inbit << (Bits - 1));
	         PulseOut(Cpin, 1, PULSECNT);
            Delay(PULSECNT);
	      }
	      break;
      case LSBPOST:
	      for(i = 0; i < Bits; i++)
	      {	 
	         PulseOut(Cpin, 1, PULSECNT);
            Inbit = In(Dpin);
	         Value = (Value >> 1) | (Inbit << (Bits - 1));
            Delay( PULSECNT);
	      }
	      break; 
   }
	return Value;
}

//----------------------------------------------------------------------------
//  i2cStop()
//----------------------------------------------------------------------------
static void i2cStop( void )
{
	 DIRA &= ~i2cClk;
    DIRA &= ~i2cDat;
}

//----------------------------------------------------------------------------
//  i2cStart()
//----------------------------------------------------------------------------
static void i2cStart( void )
{
	 DIRA &= ~i2cDat;   
    DIRA &= ~i2cClk;
    DIRA |= i2cDat;       
   
    for(;;)
    {
        if ( INA & i2cClk ) break;
    }

}

//----------------------------------------------------------------------------
//  I2C_Init()
//
//      Store I2C pins and initialize them.
//----------------------------------------------------------------------------
void I2C_Init( unsigned char pinDat, unsigned char pinClk )
{
	
  	 i2cDat = 1 << pinDat;       // Propeller pin used for I2C Data
    i2cClk = 1 << pinClk;       // Propeller pin used for I2C Clock	 
	 	
    DIRA &= ~i2cDat;            // ensure Data Pin is input to start
    OUTA &= ~i2cDat;            //      - set Dat pin Output Latch to LOW    
    DIRA &= ~i2cClk;            // ensure Clock Pin is input to start
    OUTA &= ~i2cClk;            //      - set Clock pin Output Latch to LOW
}

//----------------------------------------------------------------------------
//  I2C_IsDevPresent()
//
//      Send i2c address to check for Ack bit.
//
//      returns 0 if no device at given address
//              1 if there is a device at the given address
//----------------------------------------------------------------------------
unsigned char I2C_IsDevPresent( unsigned char devAddr )
{
    unsigned char ackbit;

    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    i2cStop();
    
    return( !ackbit );
}
    
      

//----------------------------------------------------------------------------
//  i2cWrite()
//
//      write a bite
//----------------------------------------------------------------------------
static unsigned char i2cWrite( unsigned char dat )
{
    unsigned char ackbit, idx, bit;
                                     
    DIRA |= i2cClk;             // init the clock line
 
    bit = 0x80;
 
    for( idx = 0; idx < 8; idx++ )
    { 
        if ( ( bit & dat ) == 0 )       // set the SDA while the SCL is LOW
            DIRA |= i2cDat;
        else
            DIRA &= ~i2cDat;
                       
        DIRA &= ~i2cClk;                // toggle SCL HIGH
        bit >>= 1;                      // slight delay too ...
        DIRA |= i2cClk;                 // toogle SCL LOW        
    }
       
    DIRA &= ~i2cDat;                    // setup for ACK - pin to input
                  
    DIRA &= ~i2cClk;                    // read in the ACK
    ackbit = ( ( INA & i2cDat ) != 0 );
    DIRA |= i2cClk;
     
    DIRA |= i2cDat;                     // leave the SDA pin LOW

    return( ackbit );  
}

    
//----------------------------------------------------------------------------
//  i2cRead()
//
//      read a byte
//----------------------------------------------------------------------------
static unsigned char i2cRead( unsigned char ackbit )
{
    unsigned char i2cData, idx;
    
    DIRA &= ~i2cDat;                    // set the SCL to output and the SDA to input
    DIRA |= i2cClk;

    i2cData = 0;                        // clock in the byte
    
    for ( idx = 0; idx < 8; idx++ )
    {
        i2cData <<= 1;
        DIRA &= ~i2cClk;
        if ( INA & i2cDat )
            i2cData |= 1;
        DIRA |= i2cClk;
    }
          
    DIRA |= i2cDat;                     // send the ACK or NAK
    DIRA &= ~i2cClk;
    if ( ackbit )
        DIRA &= ~i2cDat;
    else
        DIRA |= i2cDat;
        
    DIRA |= i2cClk;
    
    return( i2cData );                  // return the data
}


//----------------------------------------------------------------------------
//  I2C_ReadAddress()
//
//      Send i2c address, register address.
//      Then read a byte.
//----------------------------------------------------------------------------
unsigned char I2C_ReadAddrPage( unsigned char devAddr, unsigned int reg, unsigned char *buffAddr, unsigned int len )
{
    unsigned char ackbit, highaddr, lowaddr;
	 unsigned int temp, idx;

	 temp = reg/256;
	 highaddr = (unsigned char)temp;
	 temp = reg - (temp * 256);
	 lowaddr = (unsigned char)temp;
  
    ackbit = _i2cACK;

    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );   
    ackbit = ( ackbit << 1 ) | i2cWrite( highaddr ); 
    ackbit = ( ackbit << 1 ) | i2cWrite( lowaddr ); 

    Delay( 1600000 );            // 20mS delay (for Matrix Orbital LK-25) TODO: optimize this
  
    i2cStart();
	 ackbit = ( ackbit << 1 ) | i2cWrite( devAddr | 1 );     // repeat with read bit now set
	 for(idx = 0; idx < len; idx++)
	 {
      if (idx < len - 1)
		{
			*buffAddr++ = i2cRead( _i2cACK );
		}
		else
		{
      	*buffAddr++ = i2cRead( _i2cNAK );
		}
    }
    i2cStop();
    
    return( ackbit );
}

unsigned char I2C_ReadRegPage( unsigned char devAddr, unsigned char reg, unsigned char *buffAddr, unsigned int len )
{
    unsigned char ackbit;
	 unsigned int temp, idx;
  
    ackbit = _i2cACK;

    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );   
    ackbit = ( ackbit << 1 ) | i2cWrite( reg ); 

    Delay( 1600000 );            // 20mS delay (for Matrix Orbital LK-25) TODO: optimize this
  
    i2cStart();
	 ackbit = ( ackbit << 1 ) | i2cWrite( devAddr | 1 );     // repeat with read bit now set
	 for(idx = 0; idx < len; idx++)
	 {
      if (idx < len - 1)
		{
			*buffAddr++ = i2cRead( _i2cACK );
		}
		else
		{
      	*buffAddr++ = i2cRead( _i2cNAK );
		}
    }
    i2cStop();
    
    return( ackbit );
}

unsigned char I2C_ReadPage( unsigned char devAddr, unsigned char *buffAddr, unsigned int len )
{
    unsigned char ackbit;
	 unsigned int temp, idx;
  
    ackbit = _i2cACK;

    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );   

    Delay( 1600000 );            // 20mS delay (for Matrix Orbital LK-25) TODO: optimize this
  
    i2cStart();
	 ackbit = ( ackbit << 1 ) | i2cWrite( devAddr | 1 );     // repeat with read bit now set
	 for(idx = 0; idx < len; idx++)
	 {
      if (idx < len - 1)
		{
			*buffAddr++ = i2cRead( _i2cACK );
		}
		else
		{
      	*buffAddr++ = i2cRead( _i2cNAK );
		}
    }
    i2cStop();
    
    return( ackbit );
}


unsigned char I2C_Test( unsigned char devAddr)
{
	 return I2C_IsDevPresent( devAddr );
}

//----------------------------------------------------------------------------
//  I2C_WriteDevRegBuf()
//
//      Send i2c address, the register to write to, then the number of bytes 
//      from the specified butter by the first byte in the buffer. Returns the
//      ACK bit from the device (i.e. 1st byte = # of bytes to write).
//----------------------------------------------------------------------------
unsigned char I2C_WriteAddrPage(unsigned char devAddr, unsigned int reg, unsigned char *bufAddr, unsigned int len)
{
    unsigned char ackbit, highAddr, lowAddr;
	 unsigned int temp, idx;

	 temp = reg/256;
	 highAddr = (unsigned char)temp;
	 temp = reg - (temp * 256);
	 lowAddr = (unsigned char)temp;
  
    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    ackbit = ( ackbit << 1 ) | i2cWrite( highAddr );
	 ackbit = ( ackbit << 1 ) | i2cWrite( lowAddr );
    for ( idx = 0; idx < len; idx++ )
       ackbit = ( ackbit << 1 ) | i2cWrite( *bufAddr++ );
    i2cStop();
    
    return( ackbit );
}

unsigned char I2C_WriteRegPage(unsigned char devAddr, unsigned char reg, unsigned char *bufAddr, unsigned int len )
{
    unsigned char ackbit;
	 unsigned int temp, idx;

    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    ackbit = ( ackbit << 1 ) | i2cWrite( reg );
    for ( idx = 0; idx <len; idx++ )
       ackbit = ( ackbit << 1 ) | i2cWrite( *bufAddr++ );
    i2cStop();
    
    return( ackbit );
}

unsigned char I2C_WritePage(unsigned char devAddr, unsigned char *bufAddr, unsigned int len )
{
    unsigned char ackbit;
	 unsigned int temp, idx;

    ackbit = _i2cACK;
    
    i2cStart();
    ackbit = ( ackbit << 1 ) | i2cWrite( devAddr );
    for ( idx = 0; idx <len; idx++ )
       ackbit = ( ackbit << 1 ) | i2cWrite( *bufAddr++ );
    i2cStop();
    
    return( ackbit );
}

void SPI_init(unsigned int spi_clk, unsigned int spi_mosi, unsigned int spi_miso, unsigned char spi_mode,unsigned spi_speed)
{
	spiCLK = spi_clk;
  	spiMOSI = spi_mosi;
  	spiMISO = spi_miso;
  	spiMODE = spi_mode;
  	spiSPEED = spi_speed;

  	switch (spiMODE)
  	{
		case 0:
			Out(spiCLK, 0);
			break;
		case 1:
			Out(spiCLK, 0);
			break;
		case 2:
			Out(spiCLK, 1);
			break;
		case 3:
			Out(spiCLK, 1);
			break;
	}
}

unsigned int SPI ( unsigned int value, unsigned int bits )
{
	unsigned int i;

	switch (spiMODE)
	{
		case 0:
			for (i = 0; i < bits; i++)
			{
				Out(spiMOSI, BitRead(value, bits-1));
				value <<= 1;
				Delay(spiSPEED/2);
				Out(spiCLK, 1);
				Delay(spiSPEED/2);
				Out(spiCLK, 0);
				value |= In(spiMISO);
			}
		break;
		case 1:
			for (i = 0; i < bits; i++)
			{
				Out(spiMOSI, BitRead(value, bits-1));
				value <<= 1;
				Delay(spiSPEED/2);
				Out(spiCLK, 1);
				value |= In(spiMISO);
				Delay(spiSPEED/2);
				Out(spiCLK, 0);
			}
		break;
		case 2:
			for (i = 0; i < bits; i++)
			{
				Out(spiMOSI, BitRead(value, bits-1));
				value <<= 1;
				Delay(spiSPEED/2);
				Out(spiCLK, 0);
				Delay(spiSPEED/2);
				Out(spiCLK, 1);
				value |= In(spiMISO);
			}
		break;
		case 2:
			for (i = 0; i < bits; i++)
			{
				Out(spiMOSI, BitRead(value, bits-1));
				value <<= 1;
				Delay(spiSPEED/2);
				Out(spiCLK, 0);
				value |= In(spiMISO);
				Delay(spiSPEED/2);
				Out(spiCLK, 1);
			}
		break;
	}
